#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <process.h>
#include <rena_api.h>
#include "mana_model.h"

#define MANA_MODEL_STACK  (1024 * 1024)

//Global data
static HWND g_hwnd = NULL;
static HANDLE g_thread = INVALID_HANDLE_VALUE;
static CRITICAL_SECTION g_lock;
static bool g_end = false;
static vector<MANA_LOG> g_logs;

//Entry point
static void mana_model_entry(void *param) {
	char tag[RENA_LENGTH_TAG + 1];
	char msg[RENA_LENGTH_MSG + 1];
	unsigned long tick;
	int level;
	unsigned int start;
	unsigned int count;
	MANA_LOG *p_data;

	//Receive log loop
	while (g_end == false) {
		::mana_model_lock();

		//Start receive log
		start = g_logs.size();
		count = 0;
		while (::rena_recv(&tick, &level, tag, msg) == RENA_OK) {
			g_logs.push_back(MANA_LOG());

			//Make data
			p_data = &(*g_logs.rbegin());
			p_data->tick = tick;
			p_data->level = level;
			p_data->tag = tag;
			p_data->msg = msg;
			count++;
		}

		::mana_model_unlock();

		//Notify if received log
		if (count > 0) {
			::PostMessage(g_hwnd, MANA_WM_RECV_LOG, start, count);
		}

		//Sleep
		::Sleep(500);
	}

	//HARAKIRI
	_endthread();
}

//Initialize model
void mana_model_init(HWND hwnd) {
	::InitializeCriticalSection(&g_lock);

	//Start model thread
	g_thread = (HANDLE)_beginthread(mana_model_entry, MANA_MODEL_STACK, NULL);

	//Initialize global data
	g_hwnd = hwnd;
}

//Finalize model
void mana_model_final() {
	//End thread
	g_end = true;
	::WaitForSingleObject(g_thread, INFINITE);
}

//Lock
void mana_model_lock(void) {
	::EnterCriticalSection(&g_lock);
}

//Unlock
void mana_model_unlock(void) {
	::LeaveCriticalSection(&g_lock);
}

//Clear all logs
void mana_model_clear_logs() {
	::mana_model_lock();
	g_logs.clear();
	::mana_model_unlock();
}

//Get log list by multi condition
void mana_model_get_logs(unsigned int start, unsigned int count, int level, const string &tag, const string &filter, vector<MANA_LOG*> &logs) {
	//Adjust count
	if (count == 0xffffffff) {
		count = g_logs.size() - start;
	}

	//Make list loop
	for (unsigned int i = start; i < (start + count); i++) {
		MANA_LOG *it = &g_logs[i];

		//Check level
		if (it->level > level) {
			continue;
		}

		//Check tag
		if ((tag.length() > 0) && (it->tag != tag)) {
			continue;
		}

		//Check filter
		if ((filter.length() > 0) && (it->msg.find(filter) == string::npos)) {
			continue;
		}

		//Matched
		logs.push_back(it);
	}
}
