#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <commctrl.h>
#include <vector>
#include <string>
#include <cstring>
#include <rena_api.h>
#include "mana_model.h"
#include "resource.h"

using namespace std;

#define MANA_INI_FILE     ".\\mana.ini"
#define MANA_INI_SECTION  "mana"

//Global data
static HWND g_ctrl_run = NULL;
static HWND g_ctrl_pause = NULL;
static HWND g_ctrl_verbose = NULL;
static HWND g_ctrl_debug = NULL;
static HWND g_ctrl_info = NULL;
static HWND g_ctrl_warning = NULL;
static HWND g_ctrl_error = NULL;
static HWND g_ctrl_auto_scr = NULL;
static HWND g_ctrl_top_most = NULL;
static HWND g_ctrl_clear = NULL;
static HWND g_ctrl_list = NULL;
static HWND g_ctrl_label_tag = NULL;
static HWND g_ctrl_tag = NULL;
static HWND g_ctrl_label_filter = NULL;
static HWND g_ctrl_filter = NULL;
static HWND g_ctrl_apply = NULL;
static bool g_init = false;
static int g_level = RENA_LEVEL_VERBOSE;
static int g_pause = RENA_RUNNING;
static int g_last_x = 0;
static int g_last_y = 0;
static int g_last_width = 0;
static int g_last_height = 0;
static string g_tag = "";
static string g_filter = "";
static vector<MANA_LOG> g_logs;

//Read INI to int
int mana_read_int(const char *p_key, int def = 0) {
	return (int)::GetPrivateProfileInt(MANA_INI_SECTION, p_key, def, MANA_INI_FILE);
}

//Read INI to string
string mana_read_string(const char *p_key, const string &def = "") {
	char buffer[512];
	::GetPrivateProfileString(MANA_INI_SECTION, p_key, def.c_str(), buffer, sizeof(buffer), MANA_INI_FILE);
	return buffer;
}

//Write INI from int
void mana_write_int(const char *p_key, int value) {
	char buffer[64];
	sprintf(buffer, "%d", value);
	::WritePrivateProfileString(MANA_INI_SECTION, p_key, buffer, MANA_INI_FILE);
}

//Write INI from string
void mana_write_string(const char *p_key, const string &value) {
	::WritePrivateProfileString(MANA_INI_SECTION, p_key, value.c_str(), MANA_INI_FILE);
}

//Make listview column
void mana_make_column(const char *p_width_name, int def_width, const char *p_label, int index) {
	LV_COLUMN col;

	col.mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT | LVCF_SUBITEM;
	col.fmt = LVCFMT_LEFT;
	col.cx = ::mana_read_int(p_width_name, def_width);
	col.pszText = const_cast<char*>(p_label);
	col.iSubItem = index;
	ListView_InsertColumn(g_ctrl_list, index, &col);
}

//Set top most state
void mana_update_top_most(HWND hwnd) {
	if (::SendMessage(g_ctrl_top_most, BM_GETCHECK, 0, 0) == BST_CHECKED) {
		::SetWindowPos(hwnd, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
	}
	else {
		::SetWindowPos(hwnd, HWND_NOTOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
	}
}

//Handle WM_INITDIALOG
void mana_on_init_dialog(HWND hwnd) {
	//Initialize controls
	g_ctrl_run = ::GetDlgItem(hwnd, IDC_MAIN_RADIO_RUN);
	g_ctrl_pause = ::GetDlgItem(hwnd, IDC_MAIN_RADIO_PAUSE);
	g_ctrl_verbose = ::GetDlgItem(hwnd, IDC_MAIN_RADIO_VERBOSE);
	g_ctrl_debug = ::GetDlgItem(hwnd, IDC_MAIN_RADIO_DEBUG);
	g_ctrl_info = ::GetDlgItem(hwnd, IDC_MAIN_RADIO_INFO);
	g_ctrl_warning = ::GetDlgItem(hwnd, IDC_MAIN_RADIO_WARNING);
	g_ctrl_error = ::GetDlgItem(hwnd, IDC_MAIN_RADIO_ERROR);
	g_ctrl_auto_scr = ::GetDlgItem(hwnd, IDC_MAIN_CHECK_AUTO_SCROLL);
	g_ctrl_top_most = ::GetDlgItem(hwnd, IDC_MAIN_CHECK_TOP_MOST);
	g_ctrl_clear = ::GetDlgItem(hwnd, IDC_MAIN_BUTTON_CLEAR);
	g_ctrl_list = ::GetDlgItem(hwnd, IDC_MAIN_LIST_LOG);
	g_ctrl_label_tag = ::GetDlgItem(hwnd, IDC_MAIN_STATIC_TAG);
	g_ctrl_tag = ::GetDlgItem(hwnd, IDC_MAIN_EDIT_TAG);
	g_ctrl_label_filter = ::GetDlgItem(hwnd, IDC_MAIN_STATIC_FILTER);
	g_ctrl_filter = ::GetDlgItem(hwnd, IDC_MAIN_EDIT_FILTER);
	g_ctrl_apply = ::GetDlgItem(hwnd, IDC_MAIN_BUTTON_APPLY);

	//Initialize window position & size
	g_last_x = ::mana_read_int("x");
	g_last_y = ::mana_read_int("y");
	g_last_width = ::mana_read_int("width", 600);
	g_last_height = ::mana_read_int("height", 400);
	::SetWindowPos(hwnd, NULL, g_last_x, g_last_y, g_last_width, g_last_height, SWP_NOZORDER);

	//Initialize window type
	int type = ::mana_read_int("type", SW_SHOWNORMAL);
	::ShowWindow(hwnd, type);

	//Initialize running state
	g_pause = ::mana_read_int("pause", RENA_RUNNING);
	switch (g_pause) {
		case RENA_PAUSING:
			::SendMessage(g_ctrl_pause, BM_SETCHECK, BST_CHECKED, 0);
		case RENA_RUNNING:
		default:
			::SendMessage(g_ctrl_run, BM_SETCHECK, BST_CHECKED, 0);
			break;
	}

	//Initialize level
	g_level = ::mana_read_int("level", RENA_LEVEL_VERBOSE);
	switch (g_level) {
		case RENA_LEVEL_ERROR:
			::SendMessage(g_ctrl_error, BM_SETCHECK, BST_CHECKED, 0);
			break;
		case RENA_LEVEL_WARNING:
			::SendMessage(g_ctrl_warning, BM_SETCHECK, BST_CHECKED, 0);
			break;
		case RENA_LEVEL_INFO:
			::SendMessage(g_ctrl_info, BM_SETCHECK, BST_CHECKED, 0);
			break;
		case RENA_LEVEL_DEBUG:
			::SendMessage(g_ctrl_debug, BM_SETCHECK, BST_CHECKED, 0);
			break;
		case RENA_LEVEL_VERBOSE:
		default:
			::SendMessage(g_ctrl_verbose, BM_SETCHECK, BST_CHECKED, 0);
			break;
	}

	//Initialize auto scroll
	::SendMessage(g_ctrl_auto_scr, BM_SETCHECK, ::mana_read_int("auto_scr", 1) ? BST_CHECKED : BST_UNCHECKED, 0);

	//Initialize top most
	::SendMessage(g_ctrl_top_most, BM_SETCHECK, ::mana_read_int("top_most", 1) ? BST_CHECKED : BST_UNCHECKED, 0);
	::mana_update_top_most(hwnd);

	//Initialize listview
	::mana_make_column("cx_time", 80, "Time", 0);
	::mana_make_column("cx_level", 30, "LV", 1);
	::mana_make_column("cx_tag", 130, "Tag", 2);
	::mana_make_column("cx_msg", 330, "Message", 3);
	ListView_SetExtendedListViewStyle(g_ctrl_list, LVS_EX_FULLROWSELECT);

	//Initialize tag
	g_tag = ::mana_read_string("tag", "");
	::SendMessage(g_ctrl_tag, EM_LIMITTEXT, RENA_LENGTH_TAG, 0);
	::SetWindowText(g_ctrl_tag, g_tag.c_str());

	//Initialize filter
	g_filter = ::mana_read_string("filter", "");
	::SendMessage(g_ctrl_filter, EM_LIMITTEXT, 100, 0);
	::SetWindowText(g_ctrl_filter, g_filter.c_str());

	//First focus is ...
	::SetFocus(g_ctrl_tag);

	//Start MANA model
	::mana_model_init(hwnd);

	//Complete initialize
	g_init = true;
}

//Save column width
void mana_save_column_width(const char *p_name, int index) {
	LVCOLUMN col;

	//Read width
	col.mask = LVCF_WIDTH;
	ListView_GetColumn(g_ctrl_list, index, &col);

	//Save width
	::mana_write_int(p_name, col.cx);
}

//Handle WM_CLOSE
void mana_on_close(HWND hwnd) {
	//Save window position & size
	::mana_write_int("x", g_last_x);
	::mana_write_int("y", g_last_y);
	::mana_write_int("width", g_last_width);
	::mana_write_int("height", g_last_height);

	//Save window type
	long type = ::GetWindowLong(hwnd, GWL_STYLE);
	if (type & WS_MAXIMIZE) {
		::mana_write_int("type", SW_SHOWMAXIMIZED);
	}
	else if (type & WS_MINIMIZE) {
		::mana_write_int("type", SW_SHOWMINIMIZED);
	}
	else {
		::mana_write_int("type", SW_SHOWNORMAL);
	}

	//Save auto scroll
	::mana_write_int("auto_scr", ::SendMessage(g_ctrl_auto_scr, BM_GETCHECK, 0, 0) == BST_CHECKED);

	//Save top most
	::mana_write_int("top_most", ::SendMessage(g_ctrl_top_most, BM_GETCHECK, 0, 0) == BST_CHECKED);

	//Save tag
	::mana_write_string("tag", g_tag);

	//Save filter
	::mana_write_string("filter", g_filter);

	//Save column widths
	::mana_save_column_width("cx_time", 0);
	::mana_save_column_width("cx_level", 1);
	::mana_save_column_width("cx_tag", 2);
	::mana_save_column_width("cx_msg", 3);

	//End application
	::mana_model_final();
	::EndDialog(hwnd, 0);
}

//Memory last position & size
void mana_memory_last_pos_and_size(HWND hwnd) {
	//NOP if MAXIMIZE or MINIMIZE
	long type = ::GetWindowLong(hwnd, GWL_STYLE);
	if ((type & WS_MAXIMIZE) || (type & WS_MINIMIZE)) {
		return;
	}
	
	//Memory position
	RECT rc;
	::GetWindowRect(hwnd, &rc);
	g_last_x = rc.left;
	g_last_y = rc.top;
	g_last_width = rc.right - rc.left;
	g_last_height = rc.bottom - rc.top;
}

//Handle WM_MOVE
void mana_on_move(HWND hwnd, int x, int y) {
	//Memory last state
	::mana_memory_last_pos_and_size(hwnd);
}

//Handle WM_SIZE
void mana_on_size(HWND hwnd, int type, unsigned int width, unsigned int height) {
	RECT rc;

	::GetWindowRect(hwnd, &rc);

	//Memory last state
	::mana_memory_last_pos_and_size(hwnd);

	//Move & Resize inner controls
	if ((type == SIZE_RESTORED) || (type == SIZE_MAXIMIZED) || (type == SIZE_MAXSHOW)) {
		//ListView for log
		::SetWindowPos(g_ctrl_list, NULL, 4, 40, width - 8, height - 66, SWP_NOZORDER);

		//Clear button
		::SetWindowPos(g_ctrl_clear, NULL, width - 30, 9, 26, 23, SWP_NOSIZE | SWP_NOZORDER);

		//Tag
		::SetWindowPos(g_ctrl_label_tag, NULL, 4, height - 19, 0, 0, SWP_NOSIZE | SWP_NOZORDER);
		::SetWindowPos(g_ctrl_tag, NULL, 32, height - 22, 130, 18, SWP_NOZORDER);

		//Filter
		::SetWindowPos(g_ctrl_label_filter, NULL, 184, height - 19, 0, 0, SWP_NOSIZE | SWP_NOZORDER);
		::SetWindowPos(g_ctrl_filter, NULL, 230, height - 22, width - 320, 18, SWP_NOZORDER);

		//Apply button
		::SetWindowPos(g_ctrl_apply, NULL, width - 76, height - 23, 72, 20, SWP_NOZORDER);
	}
}

//Handle WM_GETMINMAXINFO
void mana_on_minmaxinfo(HWND hwnd, MINMAXINFO *p_mmi) {
	p_mmi->ptMinTrackSize.x = 450;
	p_mmi->ptMinTrackSize.y = 160;
}

//Update list
void mana_update_list(HWND hwnd, unsigned int start, unsigned int count) {
	vector<MANA_LOG*> logs;
	LV_ITEM item;
	char buffer[512];

	//Lock model
	::mana_model_lock();

	//Make matched list
	::mana_model_get_logs(start, count, g_level, g_tag, g_filter, logs);

	//Add to list
	::SendMessage(g_ctrl_list, WM_SETREDRAW, FALSE, 0);
	for (unsigned int i = 0; i < logs.size(); i++) {
		MANA_LOG *p_log = logs[i];
		int index;

		//Time
		sprintf(buffer, "%010u", p_log->tick);
		item.mask = LVIF_TEXT;
		item.pszText = buffer;
		item.iItem = ListView_GetItemCount(g_ctrl_list);
		item.iSubItem = 0;
		index = ListView_InsertItem(g_ctrl_list, &item);

		//Level
		switch (p_log->level) {
			case RENA_LEVEL_VERBOSE: strcpy(buffer, "V"); break;
			case RENA_LEVEL_DEBUG: strcpy(buffer, "D"); break;
			case RENA_LEVEL_INFO: strcpy(buffer, "I"); break;
			case RENA_LEVEL_WARNING: strcpy(buffer, "W"); break;
			case RENA_LEVEL_ERROR: strcpy(buffer, "E"); break;
			default: strcpy(buffer, "?"); break;
		}
		item.iItem = index;
		item.iSubItem = 1;
		ListView_SetItem(g_ctrl_list, &item);

		//Tag
		strcpy(buffer, p_log->tag.c_str());
		item.iSubItem = 2;
		ListView_SetItem(g_ctrl_list, &item);

		//Message
		strcpy(buffer, p_log->msg.c_str());
		item.iSubItem = 3;
		ListView_SetItem(g_ctrl_list, &item);
	}

	//If allow auto scroll, Select bottom
	if (::SendMessage(g_ctrl_auto_scr, BM_GETCHECK, 0, 0) == BST_CHECKED) {
		int index = ListView_GetItemCount(g_ctrl_list) - 1;
		ListView_SetItemState(g_ctrl_list, index, LVIS_SELECTED | LVIS_FOCUSED, LVIS_SELECTED | LVIS_FOCUSED);
		ListView_EnsureVisible(g_ctrl_list, index, FALSE);
	}

	//Redraw
	::SendMessage(g_ctrl_list, WM_SETREDRAW, TRUE, 0);
	::InvalidateRect(g_ctrl_list, NULL, FALSE);
	::UpdateWindow(g_ctrl_list);

	//Unlock model
	::mana_model_unlock();
}

//Changed pause state
void mana_on_change_pause(HWND hwnd, int pause) {
	g_pause = pause;
	::rena_set_pause(pause);
}

//Changed level
void mana_on_change_level(HWND hwnd, int level) {
	g_level = level;

	//Update list
	ListView_DeleteAllItems(g_ctrl_list);
	::mana_update_list(hwnd, 0, 0xffffffff);
}

//Clear log button clicked
void mana_on_click_clear(HWND hwnd) {
	//Clear log cache
	::mana_model_clear_logs();

	//Clear listview
	::SendMessage(g_ctrl_list, WM_SETREDRAW, FALSE, 0);
	ListView_DeleteAllItems(g_ctrl_list);
	::SendMessage(g_ctrl_list, WM_SETREDRAW, TRUE, 0);
	::InvalidateRect(g_ctrl_list, NULL, FALSE);
	::UpdateWindow(g_ctrl_list);
}

//Apply condition button clicked
void mana_on_click_apply(HWND hwnd) {
	char buffer[1024];

	//Memory tag
	::GetWindowText(g_ctrl_tag, buffer, sizeof(buffer));
	g_tag = buffer;

	//Memory filter
	::GetWindowText(g_ctrl_filter, buffer, sizeof(buffer));
	g_filter = buffer;

	//Update list
	ListView_DeleteAllItems(g_ctrl_list);
	::mana_update_list(hwnd, 0, 0xffffffff);
}

//Received log
void mana_on_recv_log(HWND hwnd, unsigned int start, unsigned int count) {
	::mana_update_list(hwnd, start, count);
}

//Message handler
INT_PTR CALLBACK mana_handle_event(HWND hwnd, UINT msg, WPARAM wp, LPARAM lp) {
	//Handle messages
	switch (msg) {
		case WM_INITDIALOG:
			::mana_on_init_dialog(hwnd);
			return FALSE;
		case WM_CLOSE:
			::mana_on_close(hwnd);
			return TRUE;
		case WM_MOVE:
			::mana_on_move(hwnd, (short)LOWORD(lp), (short)HIWORD(lp));
			return TRUE;
		case WM_SIZE:
			::mana_on_size(hwnd, (int)wp, LOWORD(lp), HIWORD(lp));
			return TRUE;
		case WM_GETMINMAXINFO:
			::mana_on_minmaxinfo(hwnd, (MINMAXINFO*)lp);
			return TRUE;
		case WM_COMMAND:
			//Handle click event
			if ((HIWORD(wp) == BN_CLICKED) && (g_init == true)) {
				switch (LOWORD(wp)) {
					case IDC_MAIN_RADIO_RUN:
						::mana_on_change_pause(hwnd, RENA_RUNNING);
						return TRUE;
					case IDC_MAIN_RADIO_PAUSE:
						::mana_on_change_pause(hwnd, RENA_PAUSING);
						return TRUE;
					case IDC_MAIN_RADIO_VERBOSE:
						::mana_on_change_level(hwnd, RENA_LEVEL_VERBOSE);
						return TRUE;
					case IDC_MAIN_RADIO_DEBUG:
						::mana_on_change_level(hwnd, RENA_LEVEL_DEBUG);
						return TRUE;
					case IDC_MAIN_RADIO_INFO:
						::mana_on_change_level(hwnd, RENA_LEVEL_INFO);
						return TRUE;
					case IDC_MAIN_RADIO_WARNING:
						::mana_on_change_level(hwnd, RENA_LEVEL_WARNING);
						return TRUE;
					case IDC_MAIN_RADIO_ERROR:
						::mana_on_change_level(hwnd, RENA_LEVEL_ERROR);
						return TRUE;
					case IDC_MAIN_CHECK_TOP_MOST:
						::mana_update_top_most(hwnd);
						return TRUE;
					case IDC_MAIN_BUTTON_CLEAR:
						::mana_on_click_clear(hwnd);
						return TRUE;
					case IDC_MAIN_BUTTON_APPLY:
						::mana_on_click_apply(hwnd);
						return TRUE;
				}
			}
			return FALSE;
		case MANA_WM_RECV_LOG:
			::mana_on_recv_log(hwnd, wp, lp);
			return TRUE;
		default:
			return FALSE;
	}
}

//Entry point
int APIENTRY WinMain(HINSTANCE inst, HINSTANCE prev, LPSTR p_cmd, int show) {
	//Check multi boot
	HANDLE mutex = ::CreateMutex(NULL, FALSE, "com.lmt.windows.mana.mutex");
	if(::GetLastError() == ERROR_ALREADY_EXISTS) {
		return 0;
	}

	::InitCommonControls();
	return ::DialogBox(inst, MAKEINTRESOURCE(IDD_MAIN), NULL, mana_handle_event);
}
