#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <queue>
#include <string>
#include <cstring>
#include "yuna.h"

using namespace std;

struct YUNA_LOG_ELEM {
	unsigned long tick;
	int level;
	string tag;
	string msg;
};

//Global variables
static HANDLE g_receiver = INVALID_HANDLE_VALUE;
static HANDLE g_sender = INVALID_HANDLE_VALUE;
static string g_sender_name = "";
static queue<YUNA_LOG_ELEM> g_logs;
static CRITICAL_SECTION g_lock;

//Pop front log
bool yuna_pop_log(YUNA_LOG *p_log) {
	//Abort if log is nothing
	if (g_logs.empty()) {
		return false;
	}

	::EnterCriticalSection(&g_lock);

	//Make YUNA_LOG data
	YUNA_LOG_ELEM &e = g_logs.front();
	p_log->tick = e.tick;
	p_log->level = e.level;
	strcpy(p_log->tag, e.tag.c_str());
	strcpy(p_log->msg, e.msg.c_str());

	//Pop front
	g_logs.pop();
	::LeaveCriticalSection(&g_lock);
	return true;
}

//Received SEND_LOG_REQ
static void yuna_msg_send_log(const YUNA_LOG *p_log) {
	::EnterCriticalSection(&g_lock);

	//Add to queue
	g_logs.push(YUNA_LOG_ELEM());

	YUNA_LOG_ELEM &e = g_logs.back();
	e.tick = p_log->tick;
	e.level = p_log->level;
	e.tag = p_log->tag;
	e.msg = p_log->msg;

	::LeaveCriticalSection(&g_lock);

	//Wait cancel pause
	while (::yuna_get_pause()) {
		::Sleep(10);
	}

	//Send SEND_LOG_RES
	YUNA_RES res;
	unsigned long wrote;

	res.kind = YUNA_MSG_KIND_SEND_LOG_RES;
	res.res = YUNA_RES_OK;
	::WriteFile(g_sender, &res, sizeof(res), &wrote, NULL);
}

//Entry point
void yuna_client_entry(void *param) {
	//Initialize data
	::InitializeCriticalSection(&g_lock);

	//Create mail slot
	g_receiver = ::CreateMailslot(YUNA_CLIENT_MAILSLOT, YUNA_MESSAGE_SIZE, MAILSLOT_WAIT_FOREVER, NULL);
	if (g_receiver == INVALID_HANDLE_VALUE) {
		::MessageBox(NULL, "Failed to create client mailslot.", "Yuna", MB_OK | MB_ICONSTOP);
		return;
	}

	//Wait any message
	while (1) {
		//Receive message
		unsigned char buffer[YUNA_MESSAGE_SIZE];
		unsigned long length;
		::ReadFile(g_receiver, buffer, sizeof(buffer), &length, NULL);

		//Parse message
		YUNA_MSG *p_msg = (YUNA_MSG*)buffer;
		switch (p_msg->kind) {
			case YUNA_MSG_KIND_SEND_LOG_REQ:
				::yuna_msg_send_log((YUNA_LOG*)p_msg->info);
				break;
			case YUNA_MSG_KIND_NOTIFY_MAIL:
				::yuna_notify_mail((YUNA_MAIL*)p_msg->info, &g_sender, &g_sender_name);
				break;
			default:
				continue;
		}

		//Healthy
		::yuna_accessed();
	}
}
