#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <process.h>
#include "yuna.h"

//Global data
static unsigned long g_last_access_time;
static bool g_pausing = false;

//Accessed (Report alive)
void yuna_accessed(void) {
	g_last_access_time = ::GetTickCount();
}

//Set pause flag
void yuna_set_pause(bool pause) {
	g_pausing = pause;
}

//Get pause flag
bool yuna_get_pause(void) {
	return g_pausing;
}

//Notify mail
void yuna_notify_mail(const YUNA_MAIL *p_mail, HANDLE *p_now_handle, string *p_now_name) {
	//NOP if same name
	if (*p_now_name == p_mail->name) {
		return;
	}

	//Close old mailslot
	if (*p_now_handle != INVALID_HANDLE_VALUE) {
		::CloseHandle(*p_now_handle);
	}

	//Create new sender
	*p_now_handle = ::CreateFile(p_mail->name, GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (*p_now_handle == INVALID_HANDLE_VALUE) {
		*p_now_name = "";
		return;
	}

	//Save new name
	*p_now_name = p_mail->name;
}

//Start server
static void yuna_start_server_side(void) {
	_beginthread(yuna_server_entry, YUNA_SERVER_STACK, NULL);
}

//Start client
static void yuna_start_client_side(void) {
	_beginthread(yuna_client_entry, YUNA_CLIENT_STACK, NULL);
}

//Entry point
int APIENTRY WinMain(HINSTANCE inst, HINSTANCE prev, LPSTR p_cmd, int show) {
	//Check multi boot
	HANDLE mutex = ::CreateMutex(NULL, FALSE, YUNA_MUTEX_NAME);
	if(::GetLastError() == ERROR_ALREADY_EXISTS) {
		return 0;
	}

	//Start server
	::yuna_start_server_side();

	//Start client
	::yuna_start_client_side();

	//Wait session timeout
	::yuna_accessed();
	while ((::GetTickCount() - g_last_access_time) < YUNA_ALIVE_TIME) {
		::Sleep(1000);
	}

	//SATSUGAI
	return 0;
}

