#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <string>
#include "yuna.h"

using namespace std;

//Global variables
static HANDLE g_receiver = INVALID_HANDLE_VALUE;
static HANDLE g_sender = INVALID_HANDLE_VALUE;
static string g_sender_name = "";

//Received RECV_LOG_REQ
static void yuna_msg_recv_log(void) {
	YUNA_RES res;
	YUNA_LOG *p_log = (YUNA_LOG*)res.info;
	unsigned long wrote;


	//Pop oldest log
	if (::yuna_pop_log(p_log) == false) {
		res.kind = YUNA_MSG_KIND_RECV_LOG_RES;
		res.res = YUNA_RES_NG;
	}
	else {
		res.kind = YUNA_MSG_KIND_RECV_LOG_RES;
		res.res = YUNA_RES_OK;
	}

	//Send RECV_LOG_RES
	::WriteFile(g_sender, &res, sizeof(res), &wrote, NULL);
}

//Received SET_PAUSE_REQ
static void yuna_msg_set_pause(const YUNA_PAUSE *p) {
	YUNA_RES res;
	unsigned long wrote;

	//Set pause flag
	::yuna_set_pause(p->pause ? true : false);

	//Send SET_PAUSE_RES
	res.kind = YUNA_MSG_KIND_SET_PAUSE_RES;
	res.res = YUNA_RES_OK;
	::WriteFile(g_sender, &res, sizeof(res), &wrote, NULL);
}

//Entry point
void yuna_server_entry(void *param) {
	//Create mail slot
	g_receiver = ::CreateMailslot(YUNA_SERVER_MAILSLOT, YUNA_MESSAGE_SIZE, MAILSLOT_WAIT_FOREVER, NULL);
	if (g_receiver == INVALID_HANDLE_VALUE) {
		::MessageBox(NULL, "Failed to create server mailslot.", "Yuna", MB_OK | MB_ICONSTOP);
		return;
	}

	//Wait any message
	while (1) {
		//Receive message
		unsigned char buffer[YUNA_MESSAGE_SIZE];
		unsigned long length;
		::ReadFile(g_receiver, buffer, sizeof(buffer), &length, NULL);

		//Parse message
		YUNA_MSG *p_msg = (YUNA_MSG*)buffer;
		switch (p_msg->kind) {
			case YUNA_MSG_KIND_RECV_LOG_REQ:
				::yuna_msg_recv_log();
				break;
			case YUNA_MSG_KIND_SET_PAUSE_REQ:
				::yuna_msg_set_pause((YUNA_PAUSE*)p_msg->info);
				break;
			case YUNA_MSG_KIND_NOTIFY_MAIL:
				::yuna_notify_mail((YUNA_MAIL*)p_msg->info, &g_sender, &g_sender_name);
				break;
			default:
				continue;
		}

		//Healthy
		::yuna_accessed();
	}
}
